/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS X) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "mac-prefix.h"

#include <Carbon/Carbon.h>
#include <IOKit/hid/IOHIDUsageTables.h>
#include "port.h"
#include "mac-dialog.h"
#include "mac-os.h"
#include "mac-joypad.h"
#include "HID_Utilities_External.h"

#define	kUp(i)					(i * 4)
#define	kDn(i)					(i * 4 + 1)
#define	kLf(i)					(i * 4 + 2)
#define	kRt(i)					(i * 4 + 3)

#define	kPadElemTypeNone		0
#define	kPadElemTypeHat4		1
#define	kPadElemTypeHat8		2
#define	kPadElemTypeAxis		3
#define	kPadElemTypeButton		4
#define	kPadElemTypeOtherHat4	5
#define	kPadElemTypeOtherHat8	6

#define	kPadXAxis				1
#define	kPadYAxis				0
#define	kPadHat					0

#define	kMaskUp					0x0800
#define	kMaskDn					0x0400
#define	kMaskLf					0x0200
#define	kMaskRt					0x0100

typedef struct	actionStruct
{
	pRecDevice 	fDevice;
	pRecElement	fElement;
	long		fValue;
	long		fOldValue;
}	actionRec, *actionPtr;

typedef struct	padDirectionInfo
{
	int			type;
	pRecDevice	device [2];
	pRecElement	element[2];
	long		max    [2];
	long		maxmid [2];
	long		mid    [2];
	long		midmin [2];
	long		min    [2];
}	directionInfo;

static actionRec		gActionRecs[kNeedCount];
static directionInfo	gDirectionInfo[5];

static const ControlID 	gControlIDs[kNeedCount] =
{
	{ '1_Up', 0 },
	{ '1_Dn', 0 },
	{ '1_Lf', 0 },
	{ '1_Rt', 0 },
	
	{ '2_Up', 0 },
	{ '2_Dn', 0 },
	{ '2_Lf', 0 },
	{ '2_Rt', 0 },
	
	{ '3_Up', 0 },
	{ '3_Dn', 0 },
	{ '3_Lf', 0 },
	{ '3_Rt', 0 },
	
	{ '4_Up', 0 },
	{ '4_Dn', 0 },
	{ '4_Lf', 0 },
	{ '4_Rt', 0 },
	
	{ '5_Up', 0 },
	{ '5_Dn', 0 },
	{ '5_Lf', 0 },
	{ '5_Rt', 0 },

	{ '1__B', 0 },
	{ '1__A', 0 },
	{ '1__X', 0 },
	{ '1__Y', 0 },
	{ '1__L', 0 },
	{ '1__R', 0 },
	{ '1Sel', 0 },
	{ '1Srt', 0 },
	
	{ '2__B', 0 },
	{ '2__A', 0 },
	{ '2__X', 0 },
	{ '2__Y', 0 },
	{ '2__L', 0 },
	{ '2__R', 0 },
	{ '2Sel', 0 },
	{ '2Srt', 0 },

	{ '3__B', 0 },
	{ '3__A', 0 },
	{ '3__X', 0 },
	{ '3__Y', 0 },
	{ '3__L', 0 },
	{ '3__R', 0 },
	{ '3Sel', 0 },
	{ '3Srt', 0 },

	{ '4__B', 0 },
	{ '4__A', 0 },
	{ '4__X', 0 },
	{ '4__Y', 0 },
	{ '4__L', 0 },
	{ '4__R', 0 },
	{ '4Sel', 0 },
	{ '4Srt', 0 },

	{ '5__B', 0 },
	{ '5__A', 0 },
	{ '5__X', 0 },
	{ '5__Y', 0 },
	{ '5__L', 0 },
	{ '5__R', 0 },
	{ '5Sel', 0 },
	{ '5Srt', 0 },

	{ '__FF', 0 },
	{ '_Frz', 0 },
	{ '_DeF', 0 },
	{ '_Snp', 0 },
	{ '_Esc', 0 },
	{ '_SPC', 0 },
	
	{ 'MouL', 0 },
	{ 'MouR', 0 },
	{ 'ScoT', 0 },
	{ 'ScoP', 0 },
	{ 'ScoC', 0 },
	
	{ 'Ofsc', 0 }
};

static char	gNeeds[kNeedCount][64] =
{
	"\p1P Up",
	"\p1P Down",
	"\p1P Left",
	"\p1P Right",
	
	"\p2P Up",
	"\p2P Down",
	"\p2P Left",
	"\p2P Right",
	
	"\p3P Up",
	"\p3P Down",
	"\p3P Left",
	"\p3P Right",
	
	"\p4P Up",
	"\p4P Down",
	"\p4P Left",
	"\p4P Right",

	"\p5P Up",
	"\p5P Down",
	"\p5P Left",
	"\p5P Right",
	
	"\p1P B Button",
	"\p1P A Button",
	"\p1P X Button",
	"\p1P Y Button",
	"\p1P L Button",
	"\p1P R Button",
	"\p1P Select",
	"\p1P Start",
	
	"\p2P B Button",
	"\p2P A Button",
	"\p2P X Button",
	"\p2P Y Button",
	"\p2P L Button",
	"\p2P R Button",
	"\p2P Select",
	"\p2P Start",	
	
	"\p3P B Button",
	"\p3P A Button",
	"\p3P X Button",
	"\p3P Y Button",
	"\p3P L Button",
	"\p3P R Button",
	"\p3P Select",
	"\p3P Start",	
	
	"\p4P B Button",
	"\p4P A Button",
	"\p4P X Button",
	"\p4P Y Button",
	"\p4P L Button",
	"\p4P R Button",
	"\p4P Select",
	"\p4P Start",	
	
	"\p5P B Button",
	"\p5P A Button",
	"\p5P X Button",
	"\p5P Y Button",
	"\p5P L Button",
	"\p5P R Button",
	"\p5P Select",
	"\p5P Start",	
	
	"\pFast Forward",
	"\pFreeze Game",
	"\pDefrost Game",
	"\pScreenshot",	
	"\pBreak",
	"\pSave SPC",
	
	"\pMouse Left",
	"\pMouse Right",
	"\pScope Turbo",
	"\pScope Pause",
	"\pScope Cursor",
	
	"\pOffscreen"
};

OSStatus Save_Config(void)
{
 	Boolean	syncFlag;
	short	a;
	
	JoypadSetDirectionInfo();
	
    for (a = 0; a < kNeedCount; a++)
    {
		CFStringRef	keyCFStringRef;
		char		needCStr[64], num[10];

		CopyPascalStringToC(gNeeds[a], needCStr);
		if (padSetting > 1)
		{
			sprintf(num, "_%d", padSetting);
			strcat(needCStr, num);
		}
	
		keyCFStringRef = CFStringCreateWithFormat(kCFAllocatorDefault, nil, CFSTR("%s"), needCStr);
        if (keyCFStringRef)
        {
			if (gActionRecs[a].fDevice && gActionRecs[a].fElement)
				syncFlag = HIDSaveElementPref(keyCFStringRef, kCFPreferencesCurrentApplication, gActionRecs[a].fDevice, gActionRecs[a].fElement);
            else
				CFPreferencesSetAppValue(keyCFStringRef, nil, kCFPreferencesCurrentApplication);

			CFRelease(keyCFStringRef);
        }
    }
    
	CFPreferencesAppSynchronize(kCFPreferencesCurrentApplication);

	return noErr;
}

OSStatus Load_Config(void)
{
	short	a;
	
    for (a = 0; a < kNeedCount; a++)
    {
		CFStringRef	keyCFStringRef;
		pRecDevice	pDevice  = nil;
		pRecElement	pElement = nil;
		Boolean		r = false;
		char		needCStr[64], num[10];

		CopyPascalStringToC(gNeeds[a], needCStr);
		if (padSetting > 1)
		{
			sprintf(num, "_%d", padSetting);
			strcat(needCStr, num);
		}
        
		keyCFStringRef = CFStringCreateWithFormat(kCFAllocatorDefault, nil, CFSTR("%s"), needCStr);
		if (keyCFStringRef)
		{
			r = HIDRestoreElementPref(keyCFStringRef, kCFPreferencesCurrentApplication, &pDevice, &pElement);
			if (r && pDevice && pElement)
			{	
				gActionRecs[a].fDevice  = pDevice;
				gActionRecs[a].fElement = pElement;
			}
			else
			{
				gActionRecs[a].fDevice  = nil;
				gActionRecs[a].fElement = nil;
			}
			
			CFRelease(keyCFStringRef);
		}
	}
	
	JoypadSetDirectionInfo();

	return noErr;
}

pascal OSStatus ControllerEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
	#pragma unused (inHandlerCallRef)
	
	OSStatus	err, result = eventNotHandledErr;
	WindowRef	tWindowRef;

	tWindowRef = (WindowRef) inUserData;
	
	switch (GetEventClass(inEvent))
	{
		case kEventClassWindow:
			switch (GetEventKind(inEvent))
			{
				case kEventWindowClose:
					QuitAppModalLoopForWindow(tWindowRef);
					result = noErr;
			}
			
			break;
		
		case kEventClassCommand:
			switch (GetEventKind(inEvent))
			{
				case kEventCommandProcess:
					HICommand	tHICommand;

					err = GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, nil, sizeof(HICommand), nil, &tHICommand);
					if (err == noErr)
					{
						if (tHICommand.commandID == 'CLRa')
						{
							ClearPadSetting();
							result = noErr;
						}
						else
						{	
							SInt32	command = -1, count;
							
							for (count = 0; count < kNeedCount; count++)
								if (tHICommand.commandID == gControlIDs[count].signature)
									command = count;
	
							if (command >= 0)
							{
								pRecDevice	pDevice;
								pRecElement	pElement;
								
								FlushEventQueue(GetCurrentEventQueue());
	
								if (HIDConfigureAction(&pDevice, &pElement, 2.5))
								{
									if (command < 20)	// Direction
									{
										int	i = command >> 2;	// Player
										
										if (pElement->usage == kHIDUsage_GD_Hatswitch)	// Hat Switch
										{
											gActionRecs[kUp(i)].fDevice  = gActionRecs[kDn(i)].fDevice  = gActionRecs[kLf(i)].fDevice  = gActionRecs[kRt(i)].fDevice  = pDevice;
											gActionRecs[kUp(i)].fElement = gActionRecs[kDn(i)].fElement = gActionRecs[kLf(i)].fElement = gActionRecs[kRt(i)].fElement = pElement;
										
											if (pDevice->vendorID == 1103)	// Thrustmaster
												gDirectionInfo[i].type = (pElement->max > 4) ? kPadElemTypeOtherHat8 : kPadElemTypeOtherHat4;			
											else
												gDirectionInfo[i].type = (pElement->max > 4) ? kPadElemTypeHat8      : kPadElemTypeHat4;		
									
											gDirectionInfo[i].device [kPadHat] = pDevice;
											gDirectionInfo[i].element[kPadHat] = pElement;
											gDirectionInfo[i].max    [kPadHat] = pElement->max;
											gDirectionInfo[i].min    [kPadHat] = pElement->min;							
										}
										else
										if (pElement->max - pElement->min > 1)			// Axis (maybe)
										{
											if ((command % 4 == 0) || (command % 4 == 1))	// Up or Dn
											{
												gActionRecs[kUp(i)].fDevice  = gActionRecs[kDn(i)].fDevice  = pDevice;
												gActionRecs[kUp(i)].fElement = gActionRecs[kDn(i)].fElement = pElement;
											
												gDirectionInfo[i].type               = kPadElemTypeAxis;
												gDirectionInfo[i].device [kPadYAxis] = pDevice;
												gDirectionInfo[i].element[kPadYAxis] = pElement;
												gDirectionInfo[i].max    [kPadYAxis] = pElement->max;
												gDirectionInfo[i].min    [kPadYAxis] = pElement->min;
												gDirectionInfo[i].mid    [kPadYAxis] = (gDirectionInfo[i].max[kPadYAxis] + gDirectionInfo[i].min[kPadYAxis]) >> 1;
												gDirectionInfo[i].maxmid [kPadYAxis] = (gDirectionInfo[i].max[kPadYAxis] + gDirectionInfo[i].mid[kPadYAxis]) >> 1;
												gDirectionInfo[i].midmin [kPadYAxis] = (gDirectionInfo[i].mid[kPadYAxis] + gDirectionInfo[i].min[kPadYAxis]) >> 1;
											}
											else											// Lf or Rt
											{
												gActionRecs[kLf(i)].fDevice  = gActionRecs[kRt(i)].fDevice  = pDevice;
												gActionRecs[kLf(i)].fElement = gActionRecs[kRt(i)].fElement = pElement;
											
												gDirectionInfo[i].type               = kPadElemTypeAxis;
												gDirectionInfo[i].device [kPadXAxis] = pDevice;
												gDirectionInfo[i].element[kPadXAxis] = pElement;
												gDirectionInfo[i].max    [kPadXAxis] = pElement->max;
												gDirectionInfo[i].min    [kPadXAxis] = pElement->min;
												gDirectionInfo[i].mid    [kPadXAxis] = (gDirectionInfo[i].max[kPadXAxis] + gDirectionInfo[i].min[kPadXAxis]) >> 1;
												gDirectionInfo[i].maxmid [kPadXAxis] = (gDirectionInfo[i].max[kPadXAxis] + gDirectionInfo[i].mid[kPadXAxis]) >> 1;
												gDirectionInfo[i].midmin [kPadXAxis] = (gDirectionInfo[i].mid[kPadXAxis] + gDirectionInfo[i].min[kPadXAxis]) >> 1;
											}										
										}
										else											// Button (maybe)
										{
											gActionRecs[command].fDevice  = pDevice;
											gActionRecs[command].fElement = pElement;
											gDirectionInfo[i].type = kPadElemTypeButton;
										}
									}
									else
									{
										gActionRecs[command].fDevice  = pDevice;
										gActionRecs[command].fElement = pElement;							
									}
								}
								else
								{
									if (command < 20)	// Direction
									{
										int	i = command >> 2;	// Player
										
										gActionRecs[kUp(i)].fDevice  = gActionRecs[kDn(i)].fDevice  = gActionRecs[kLf(i)].fDevice  = gActionRecs[kRt(i)].fDevice  = nil;
										gActionRecs[kUp(i)].fElement = gActionRecs[kDn(i)].fElement = gActionRecs[kLf(i)].fElement = gActionRecs[kRt(i)].fElement = nil;
										
										gDirectionInfo[i].type = kPadElemTypeNone;
										gDirectionInfo[i].device [0] = gDirectionInfo[i].device [1] = nil;
										gDirectionInfo[i].element[0] = gDirectionInfo[i].element[1] = nil;
									}
									else
									{
										gActionRecs[command].fDevice  = nil;
										gActionRecs[command].fElement = nil;
									}
								}
								
								gActionRecs[command].fValue    = 0;
								gActionRecs[command].fOldValue = -2;
								
								FlushEventQueue(GetCurrentEventQueue());
								
								result = noErr;
							}
						}
					}
			}
	}
		
	return result;
}

static pascal void IdleTimer(EventLoopTimerRef inTimer, void *userData)
{
	#pragma unused (inTimer)

	WindowRef	window = (WindowRef) userData;
	ControlRef	control;
 	int			count, i;
	uint32		pad[5];
	
	static uint32	old[5] = { ~0, ~0, ~0, ~0, ~0 };
	
	for (i = 0; i < 5; i++)
	{
		pad[i] = 0;
		
		JoypadScanDirection(i, &(pad[i]));
		if (pad[i] != old[i])
		{
			old[i] = pad[i];
			
			GetControlByID(window, &gControlIDs[kUp(i)], &control);
			SetControl32BitValue(control, (pad[i] & kMaskUp) ? 1 : 0);
			GetControlByID(window, &gControlIDs[kDn(i)], &control);
			SetControl32BitValue(control, (pad[i] & kMaskDn) ? 1 : 0);
			GetControlByID(window, &gControlIDs[kLf(i)], &control);
			SetControl32BitValue(control, (pad[i] & kMaskLf) ? 1 : 0);
			GetControlByID(window, &gControlIDs[kRt(i)], &control);
			SetControl32BitValue(control, (pad[i] & kMaskRt) ? 1 : 0);
		}
	}

	for (count = 20; count < kNeedCount; count++)
	{
		gActionRecs[count].fValue = ISpKeyIsPressed(count);

		if (gActionRecs[count].fOldValue != gActionRecs[count].fValue)
		{
			GetControlByID(window, &gControlIDs[count], &control);
			SetControl32BitValue(control, (gActionRecs[count].fValue ? 1 : 0));

			gActionRecs[count].fOldValue  = gActionRecs[count].fValue;
		}
	}
}

void SetUpHID(void)
{	
	pRecDevice	device;
	
	HIDBuildDeviceList(nil, nil);
	device = HIDGetFirstDevice();
	if (!device)
	{
		hidExist = false;
		return;
	}
	
	hidExist = true;
	
	ClearPadSetting();

	Load_Config();
}

void ClearPadSetting(void)
{
	int	i;
	
	for (i = 0; i < 5; i++)
	{
		gDirectionInfo[i].type = kPadElemTypeNone;
		gDirectionInfo[i].device [0] = gDirectionInfo[i].device [1] = nil;
		gDirectionInfo[i].element[0] = gDirectionInfo[i].element[1] = nil;
	}
	
	for (i = 0; i < kNeedCount; i++)
	{
		gActionRecs[i].fDevice   = nil;
		gActionRecs[i].fElement  = nil;
		gActionRecs[i].fValue    = 0;
		gActionRecs[i].fOldValue = -2;
	}
}

void ReleaseHID(void)
{
	if (hidExist)
		HIDReleaseDeviceList();
}

void ConfigureHID(void)
{
	OSStatus	err;
	IBNibRef	nibRef;
	
	if (!hidExist)
		return;
	
	err = CreateNibReference(CFSTR("SNES9X"), &nibRef);
	if (err == noErr)
	{
		WindowRef	tWindowRef;
		
		err = CreateWindowFromNib(nibRef, CFSTR("Controllers"), &tWindowRef);
		if (err == noErr)
		{
			ControlRef			control;
			ControlID			controlID;
			CFStringRef			str1, str2;
			EventHandlerRef		eref;
			EventLoopTimerRef	tref;
			EventHandlerUPP		eventUPP;
			EventLoopTimerUPP	timerUPP;
			EventTypeSpec		windowEvents[] = { { kEventClassCommand, kEventCommandProcess }, { kEventClassWindow, kEventWindowClose } };

			Load_Config();
			
			controlID.id = 0;
			controlID.signature = 'PRES';
			GetControlByID(tWindowRef, &controlID, &control);
			str1 = CFCopyLocalizedString(CFSTR("PresetNum"), "PresetNum");
			str2 = CFStringCreateWithFormat(kCFAllocatorDefault, nil, str1, padSetting);
			SetStaticTextCFString(control, str2, false);
			CFRelease(str2);
			CFRelease(str1);

			eventUPP = NewEventHandlerUPP(ControllerEventHandler);
			err = InstallWindowEventHandler(tWindowRef, eventUPP, 2, windowEvents, (void *) tWindowRef, &eref);

			timerUPP = NewEventLoopTimerUPP(IdleTimer);
			err = InstallEventLoopTimer(GetCurrentEventLoop(), 0.0f, 0.1f, timerUPP, (void *) tWindowRef, &tref);

			ShowWindow(tWindowRef);
			err = RunAppModalLoopForWindow(tWindowRef);
			HideWindow(tWindowRef);
			
			err = RemoveEventLoopTimer(tref);
			DisposeEventLoopTimerUPP(timerUPP);
			
			err = RemoveEventHandler(eref);
			DisposeEventHandlerUPP(eventUPP);
			
			ReleaseWindow(tWindowRef);

			Save_Config();
		}
		
		DisposeNibReference(nibRef);
	}
}

long ISpKeyIsPressed(short needID)
{
	return (gActionRecs[needID].fDevice ? HIDGetElementValue(gActionRecs[needID].fDevice, gActionRecs[needID].fElement) : 0);
}

void JoypadScanDirection(int i, uint32 *controlPad)
{
	long	state;
	
	switch (gDirectionInfo[i].type)
	{
		case kPadElemTypeAxis:							// Axis (maybe)
			if (gDirectionInfo[i].device[kPadYAxis])	// Y-Axis
			{
				state = HIDGetElementValue(gDirectionInfo[i].device[kPadYAxis], gDirectionInfo[i].element[kPadYAxis]);			
				if (state >= gDirectionInfo[i].maxmid[kPadYAxis])
					*controlPad |= kMaskDn;
				else
				if (state <= gDirectionInfo[i].midmin[kPadYAxis])
					*controlPad |= kMaskUp;
			}
			
			if (gDirectionInfo[i].device[kPadXAxis])	// X-Axis
			{				
				state = HIDGetElementValue(gDirectionInfo[i].device[kPadXAxis], gDirectionInfo[i].element[kPadXAxis]);			
				if (state >= gDirectionInfo[i].maxmid[kPadXAxis])
					*controlPad |= kMaskRt;
				else
				if (state <= gDirectionInfo[i].midmin[kPadXAxis])
					*controlPad |= kMaskLf;
			}
			
			break;
		
		case kPadElemTypeHat8:							// Hat Switch (8 Directions)
			if (gDirectionInfo[i].device[kPadHat])
			{
				state = HIDGetElementValue(gDirectionInfo[i].device[kPadHat], gDirectionInfo[i].element[kPadHat]);
				switch (state)
				{
					case 1:	*controlPad |=  kMaskUp			  ;	break;
					case 2:	*controlPad |= (kMaskUp | kMaskRt);	break;
					case 3:	*controlPad |=  kMaskRt			  ;	break;
					case 4:	*controlPad |= (kMaskRt | kMaskDn);	break;
					case 5:	*controlPad |=  kMaskDn			  ;	break;
					case 6:	*controlPad |= (kMaskDn | kMaskLf);	break;
					case 7:	*controlPad |=  kMaskLf			  ;	break;
					case 8:	*controlPad |= (kMaskLf | kMaskUp);
				}
			}
				
			break;
		
		case kPadElemTypeHat4:							// Hat Switch (4 Directions)
			if (gDirectionInfo[i].device[kPadHat])
			{
				state = HIDGetElementValue(gDirectionInfo[i].device[kPadHat], gDirectionInfo[i].element[kPadHat]);
				switch (state)
				{
					case 1:	*controlPad |=  kMaskUp;	break;
					case 2:	*controlPad |=  kMaskRt;	break;
					case 3:	*controlPad |=  kMaskDn;	break;
					case 4:	*controlPad |=  kMaskLf;
				}
			}
			
			break;
			
		case kPadElemTypeOtherHat8:						// Hat Switch (8 Directions, Start at 0)
			if (gDirectionInfo[i].device[kPadHat])
			{
				state = HIDGetElementValue(gDirectionInfo[i].device[kPadHat], gDirectionInfo[i].element[kPadHat]);
				switch (state)
				{
					case 0:	*controlPad |=  kMaskUp			  ;	break;
					case 1:	*controlPad |= (kMaskUp | kMaskRt);	break;
					case 2:	*controlPad |=  kMaskRt			  ;	break;
					case 3:	*controlPad |= (kMaskRt | kMaskDn);	break;
					case 4:	*controlPad |=  kMaskDn			  ;	break;
					case 5:	*controlPad |= (kMaskDn | kMaskLf);	break;
					case 6:	*controlPad |=  kMaskLf			  ;	break;
					case 7:	*controlPad |= (kMaskLf | kMaskUp);
				}
			}
				
			break;
		
		case kPadElemTypeOtherHat4:						// Hat Switch (4 Directions, Start at 0)
			if (gDirectionInfo[i].device[kPadHat])
			{
				state = HIDGetElementValue(gDirectionInfo[i].device[kPadHat], gDirectionInfo[i].element[kPadHat]);
				switch (state)
				{
					case 0:	*controlPad |=  kMaskUp;	break;
					case 1:	*controlPad |=  kMaskRt;	break;
					case 2:	*controlPad |=  kMaskDn;	break;
					case 3:	*controlPad |=  kMaskLf;
				}
			}
			
			break;
			
		case kPadElemTypeButton:						// Button (maybe)
			if (gActionRecs[kUp(i)].fDevice && HIDGetElementValue(gActionRecs[kUp(i)].fDevice, gActionRecs[kUp(i)].fElement))
				*controlPad |= kMaskUp;
			if (gActionRecs[kDn(i)].fDevice && HIDGetElementValue(gActionRecs[kDn(i)].fDevice, gActionRecs[kDn(i)].fElement))
				*controlPad |= kMaskDn;
			if (gActionRecs[kLf(i)].fDevice && HIDGetElementValue(gActionRecs[kLf(i)].fDevice, gActionRecs[kLf(i)].fElement))
				*controlPad |= kMaskLf;
			if (gActionRecs[kRt(i)].fDevice && HIDGetElementValue(gActionRecs[kRt(i)].fDevice, gActionRecs[kRt(i)].fElement))
				*controlPad |= kMaskRt;
	}
}		

void JoypadSetDirectionInfo(void)
{	
	for (int i = 0; i < 5; i++)
	{
		if (((gActionRecs[kUp(i)].fDevice) && (gActionRecs[kUp(i)].fElement)) &&
			((gActionRecs[kDn(i)].fDevice) && (gActionRecs[kDn(i)].fElement)) &&
			((gActionRecs[kLf(i)].fDevice) && (gActionRecs[kLf(i)].fElement)) &&
			((gActionRecs[kRt(i)].fDevice) && (gActionRecs[kRt(i)].fElement)))
		{
			if ((gActionRecs[kUp(i)].fDevice  == gActionRecs[kDn(i)].fDevice)  &&
				(gActionRecs[kDn(i)].fDevice  == gActionRecs[kLf(i)].fDevice)  &&
				(gActionRecs[kLf(i)].fDevice  == gActionRecs[kRt(i)].fDevice)  &&
				(gActionRecs[kUp(i)].fElement == gActionRecs[kDn(i)].fElement) &&
				(gActionRecs[kDn(i)].fElement == gActionRecs[kLf(i)].fElement) &&
				(gActionRecs[kLf(i)].fElement == gActionRecs[kRt(i)].fElement) &&
				(gActionRecs[kUp(i)].fElement->usage == kHIDUsage_GD_Hatswitch))				// Hat Switch
			{
				if (gActionRecs[kUp(i)].fDevice->vendorID == 1103) // Thrustmaster
					gDirectionInfo[i].type = (gActionRecs[kUp(i)].fElement->max > 4) ? kPadElemTypeOtherHat8 : kPadElemTypeOtherHat4;				
				else
					gDirectionInfo[i].type = (gActionRecs[kUp(i)].fElement->max > 4) ? kPadElemTypeHat8      : kPadElemTypeHat4;			
				
				gDirectionInfo[i].device [kPadHat] =  gActionRecs[kUp(i)].fDevice;
				gDirectionInfo[i].element[kPadHat] =  gActionRecs[kUp(i)].fElement;
				gDirectionInfo[i].max    [kPadHat] =  gActionRecs[kUp(i)].fElement->max;
				gDirectionInfo[i].min    [kPadHat] =  gActionRecs[kUp(i)].fElement->min;				
			}
			else
			if ((gActionRecs[kUp(i)].fDevice  == gActionRecs[kDn(i)].fDevice)               &&
				(gActionRecs[kLf(i)].fDevice  == gActionRecs[kRt(i)].fDevice)               &&
				(gActionRecs[kUp(i)].fElement == gActionRecs[kDn(i)].fElement)              &&
				(gActionRecs[kLf(i)].fElement == gActionRecs[kRt(i)].fElement)              &&
				(gActionRecs[kUp(i)].fElement->max - gActionRecs[kUp(i)].fElement->min > 1) &&
				(gActionRecs[kLf(i)].fElement->max - gActionRecs[kLf(i)].fElement->min > 1))	// Axis (maybe)
			{				
				gDirectionInfo[i].type = kPadElemTypeAxis;

				gDirectionInfo[i].device [kPadYAxis] = gActionRecs[kUp(i)].fDevice;
				gDirectionInfo[i].element[kPadYAxis] = gActionRecs[kUp(i)].fElement;			
				gDirectionInfo[i].max    [kPadYAxis] = gActionRecs[kUp(i)].fElement->max;
				gDirectionInfo[i].min    [kPadYAxis] = gActionRecs[kUp(i)].fElement->min;
				gDirectionInfo[i].mid    [kPadYAxis] = (gDirectionInfo[i].max[kPadYAxis] + gDirectionInfo[i].min[kPadYAxis]) >> 1;
				gDirectionInfo[i].maxmid [kPadYAxis] = (gDirectionInfo[i].max[kPadYAxis] + gDirectionInfo[i].mid[kPadYAxis]) >> 1;
				gDirectionInfo[i].midmin [kPadYAxis] = (gDirectionInfo[i].mid[kPadYAxis] + gDirectionInfo[i].min[kPadYAxis]) >> 1;
				
				gDirectionInfo[i].device [kPadXAxis] = gActionRecs[kLf(i)].fDevice;
				gDirectionInfo[i].element[kPadXAxis] = gActionRecs[kLf(i)].fElement;			
				gDirectionInfo[i].max    [kPadXAxis] = gActionRecs[kLf(i)].fElement->max;
				gDirectionInfo[i].min    [kPadXAxis] = gActionRecs[kLf(i)].fElement->min;
				gDirectionInfo[i].mid    [kPadXAxis] = (gDirectionInfo[i].max[kPadXAxis] + gDirectionInfo[i].min[kPadXAxis]) >> 1;
				gDirectionInfo[i].maxmid [kPadXAxis] = (gDirectionInfo[i].max[kPadXAxis] + gDirectionInfo[i].mid[kPadXAxis]) >> 1;
				gDirectionInfo[i].midmin [kPadXAxis] = (gDirectionInfo[i].mid[kPadXAxis] + gDirectionInfo[i].min[kPadXAxis]) >> 1;
			}
			else																				// Button (maybe)
				gDirectionInfo[i].type = kPadElemTypeButton;
		}
		else
		{
			gActionRecs[kUp(i)].fDevice  = gActionRecs[kDn(i)].fDevice  = gActionRecs[kLf(i)].fDevice  = gActionRecs[kRt(i)].fDevice  = nil;
			gActionRecs[kUp(i)].fElement = gActionRecs[kDn(i)].fElement = gActionRecs[kLf(i)].fElement = gActionRecs[kRt(i)].fElement = nil;

			gDirectionInfo[i].type = kPadElemTypeNone;
			gDirectionInfo[i].device [0] = gDirectionInfo[i].device [1] = nil;
			gDirectionInfo[i].element[0] = gDirectionInfo[i].element[1] = nil;	
		}
	}
}
